import std/logging, std/os, std/strutils
import zero_functional

import ./ast

type
  FormatContext = ref object
    chart: ChordChart
    currentKey: ChordChartPitch
    sourceKey: ChordChartPitch

const DEFAULT_STYLESHEET* = """
<style>
* {
  box-sizing: border-box;
  margin: 0;
  padding: 0;
}

html { font-family: sans-serif; }

h2 { font-size: 1.25em; }

section { margin-top: 1em; }

h3 {
  font-size: 1.125rem;
  margin-bottom: 0.5em;
  text-decoration: underline;
}

.line {
  display: flex;
  flex-direction: row;
  align-items: flex-end;

  margin-left: 0.5em;
  margin-bottom: 0.5em;
}

.word {
  display: flex;
  flex-direction: column;
  align-items: flex-start;
}

.word.space-after { margin-right: 0.5em; }

.chord {
  font-weight: 600;
  margin-right: 0.5em;
}

@media screen {
  body { margin: 1em; }
}
</style>
"""

proc transpose(ctx: FormatContext, chord: ChordChartChord): string =
  let distance = ctx.currentKey - ctx.sourceKey
  if distance != 0: return $(chord.rootPitch + distance) & chord.flavor
  else: return chord.original

proc toHtml(ctx: var FormatContext, node: ChordChartNode, indent: string): string =
  result = ""
  case node.kind
  of ccnkSection:
    result &= indent & "<section>\p" &
      indent & "  " & "<h3>" & node.sectionName & "</h3>\p"

    var contents = newSeq[string]()
    for contentNode in node.sectionContents:
      contents.add(ctx.toHtml(contentNode, indent & "  "))
    result &= contents.join("\p")

    result &= indent & "</section>"

  of ccnkLine:
    result &= indent & "<div class=line>\p"
    for linePart in node.line: result &= ctx.toHtml(linePart, indent & "  ")
    result &= indent & "</div>"

  of ccnkWord:
    result &= "<span class='word"
    #if node.spaceBefore: result&=" space-before"
    if node.spaceAfter: result&=" space-after"
    result &= "'>"

    if node.chord.original.len > 0:
      result &= "<span class=chord>" & ctx.transpose(node.chord) & "</span>"

    result &= "<span class=lyric>"
    if node.word.len > 0: result &= node.word
    else: result &= "&nbsp;"
    result &= "</span>"

    result &= "</span>"

  of ccnkNote:
    result &= indent & "<div class=note>" & node.note & "</div>"

  of ccnkColBreak: discard #FIXME
  of ccnkPageBreak: discard #FIXME

  of ccnkTransposeKey:
    ctx.currentKey = ctx.currentKey + node.transposeSteps

  of ccnkRedefineKey:
    ctx.sourceKey = ctx.currentKey + node.transposeSteps
    ctx.currentKey = ctx.sourceKey

proc toHtml*(
    cc: ChordChart,
    stylesheets = @[DEFAULT_STYLESHEET],
    scripts: seq[string] = @[]): string =

  var ctx = FormatContext(
    chart: cc,
    currentKey: cc.metadata.key,
    sourceKey: cc.metadata.key)

  result = """<!doctype html>
<html>
  <head>
    <title>""" & cc.metadata.title & "</title>\p"

  for ss in stylesheets:
    if ss.startsWith("<style>"): result &= ss
    elif fileExists(ss): result &= "<style>\p" & readFile(ss) & "\p</style>"
    else: warn "cannot read stylesheet file '" & ss & "'"

  for sc in scripts:
    if sc.startsWith("<script"): result &= sc
    elif fileExists(sc): result &= "<script>\p" & readFile(sc) & "\p</script>"
    else: warn "cannot read script file '" & sc & "'"

  result &= "  </head>\p  <body>"

  var indent = "      "
  #
  # Title
  result &= indent & "<h1>" & cc.metadata.title & "</h1>\p"

  result &= indent & "<h2>Key: " & $cc.metadata.key
  if cc.metadata.contains("bpm"): result &= " " & cc.metadata["bpm"] & "bpm"
  result &= "</h2>\p"

  result &= join(cc.nodes --> map(ctx.toHtml(it, indent & "  ")), "\p")

  result &= "  </body>\p</html>"
